/////////////////////////////////////////////////////////
//
// Set the positions on a surface mesh.
//
//
//
//
// The surface is represented by a quadrilateral mesh, as
// depicted below. The integer-valued array-coordinates 
// run from 0 to sDim-1 in one direction, and from 0 to tDim-1 
// in the other.
//
//  ti = tDim-1 ------------------------- . . -------  t = 1.0   y =  1.0
//              |     |     |     |     |     |     |
//              |     |     |     |     |     |     |
//              |     |     |     |     |     |     |
//              -------------------------     -------
//              .     .     .     .                 .     .        .
//             
//              .     .     .     .                 .     .        .
//              -------------------------     -------
//              |     |     |     |     |     |     |
//              |     |     |     |     |     |     |
//              |     |     |     |     |     |     |
//  ti = 3      ------------------------- . . -------
//              |     |     |     |     |     |     |
//              |     |     |     |     |     |     |
//              |     |     |     |     |     |     |
//  ti = 2      ------------------------- . . -------
//              |     |     |     |     |     |     |
//              |     |     |     |     |     |     |
//              |     |     |     |     |     |     |
//  ti = 1      ------------------------- . . -------
//              |     |     |     |     |     |     |
//              |     |     |     |     |     |     |
//              |     |     |     |     |     |     |
//  ti = 0      ------------------------- . . -------  t = 0.0   y = -1.0
//       
//       
//        si =  0     1     2     3       . .    sDim-1
//       
//        s  =  0.0                       . .     1.0
//       
//        x  = -1.0                       . .     1.0
//       
//
// Floating point parameters s and t run from 0.0 to 1.0.
//
// Each vertex of the mesh has coordinates (x,y,z), with x and y
// running from -1.0 to 1.0, defining a square in physical space.
//
// The time-angle mesh serves as an index into a texture map
// whose colors are displayed on the surface.


/////////////////////////////////////////////////////////////////////////////
//                                                                           
// Includes

#include <stdio.h>
#include <math.h>
#include <time.h>

#include "grid.h"
#include "oscillator.h"
#include "rippleTank.h"
#include "lookup.h"

/////////////////////////////////////////////////////////////////////////////
//                                                                            
// Global Variables

extern OscillatorStruct Oscillator;
extern RippleTankStruct RippleTank;
extern GridStruct       Grid;

extern float            SimulationTime;
extern double           TimeIncrement;
extern LookupTable      L;
extern int              OptUseLookups;
extern int              OptShowTimes;


/////////////////////////////////////////////////////////////////////////////
//                                                                            
//  WWWW                        WWW                                           
// WW  WW  WW                    WW            WW   WW                        
// WWW                           WW            WW                             
//  WWWW   WW  WWWW WW   WW WW   WW    WWWW   WWWW  WW   WWW   WWWW           
//    WWW  WW  WW WW WW  WW WW   WW       WW   WW   WW  WW WW  WW WW          
//     WW  WW  WW WW WW  WW WW   WW    WWWWW   WW   WW  WW WW  WW WW          
// WW  WW  WW  WW WW WW  WW WW   WW   WW  WW   WW   WW  WW WW  WW WW          
//  WWWW   WW  WW WW WW   WWWW  WWWW   WWWWW   WW   WW   WWW   WW WW          
//                                                                            
// Simulation



///////////////////////////////////////////////////////////
// 
// Simulation 
//
// Simulate an oscillator moving through a ripple tank.
//
// At time t-dt ago, the oscillator currently at point p
// was at position
//
//                  dp
//   p'  =  p - dt -----                   (1)
//                  dt
//
// Meanwhile, its wave has propagated a distance
//
//   d   =  v dt
//
// where v is the wave speed in the medium. Therefore
// the distance between p' and q is
//
//   || p' - q ||  =  v dt                 (2)
//
// The height at a point q matches the height
// at p' at time t-dt. We solve for dt, then find p',
// by substituting (1) into (2), which leads to a
// quadratic equation in dt.
//  
//   a dt^2   +   b dt   +   c   =  0
//
//
// When the magnitude of the oscillator's velocity is
// greater than the wave speed in the ripple tank
// (i.e., it exceeds Mach 1), a discontinuity forms behind the
// moving oscillator. The discontinuity occurs along the
// locus of points q for which the discriminant of the
// quadratic is zero.
//


/////////////////////////////////////////////////////////////////////////////
//                                    W     W                     W    W      
//                W         W         WW   WW              W     W      W     
//                W         W         W W W W              W     W      W     
// W  W  WWW    W W   WWW  WWW   WW   W  W  W   WW    WWW  WWW   W      W     
// W  W  W  W  W WW  W  W   W   W  W  W     W  W  W  W     W  W  W      W     
// W  W  W  W  W  W  W  W   W   WWWW  W     W  WWWW   WW   W  W  W      W     
// W  W  W  W  W  W  W WW   W   W     W     W  W        W  W  W  W      W     
//  WWW  WWW    WWW   W W   WW   WWW  W     W   WWW  WWW   W  W  W      W     
//       W                                                        W    W      
//       W                                                                    
//
// Set the positions on a surface mesh.
// Set the texture coordinates according to surface height.

void 
updateMesh(int gridpointBudget) 
  {
  int      si, ti;          // loop variables for parameters s, t
  int      sInit, tInit;    // initial values for loop variables
  double   s, t;            // parameters and increments
  double   x, y, z;         // coordinates of a vertex 
  double   distance2;       // distance^2 from oscillator
  int      index;           // sequential index into array 

  double   amplitude  = Oscillator.amplitude;
  double   frequency  = Oscillator.frequency;
  double   phase      = Oscillator.phase;
  double   waveSpeed  = RippleTank.waveSpeed;
  double   damping    = RippleTank.damping;

  double   dampingScale;
  double   angle;

  double   a, b, c;        // coefficients for quadratic equation for dt
  //double   d;              // discriminant for quadratic equation for dt
  int d;
  double   dt, dt1, dt2;   // dt1 and dt2 are the two roots of the quadratic
  SbVec2f  p, dp;          // oscillator position and directional velocity
  SbVec2f  q;              // position (projected to R2) of a grid point

  int        gridpointsCompleted = 0;
  int        breakFlag           = 0;
  static int StartingIndex       = 0;
  static int CompletedRows       = 0;
  static int oldtime = 0;


  // create drop data structure
  struct dropStruct {
    int s;
    int t;
    double time;
  };

  const int MAX_DROPS = 15;
  static dropStruct drops[MAX_DROPS];


  // initalize drops
  static int nextDrop = 0;
  static int setupDone = 0;
  if (!setupDone){
    for (int i=0; i<MAX_DROPS; i++)
      drops[i].time = -TimeIncrement;
    /*
    drops[0].s=Grid.sDim/4;
    drops[0].t=Grid.tDim/4;;
    drops[0].time = 0.0;

    drops[1].s=3*Grid.sDim/4;
    drops[1].t=3*Grid.tDim/4;;
    drops[1].time = TimeIncrement*10;

    drops[2].time = -TimeIncrement;
    drops[3].time = -TimeIncrement;
    */
    setupDone++;
  }
  

  p  = Oscillator.position;
  dp = Oscillator.velocity;

  if (drand48()>0.6){
    drops[nextDrop].s = int((Oscillator.position[0]+5.0)/10.0*Grid.sDim);
    drops[nextDrop].t = int((Oscillator.position[1]+5.0)/10.0*Grid.tDim);
    drops[nextDrop].time = SimulationTime;
  
    nextDrop++;
    if (nextDrop>=MAX_DROPS)
      nextDrop = 0;
  }

  // INVENTORism.
  //
  // In order to change the coordinates in the SoCoordinate3
  // or the SoTextureCoordinate2 array, we begin by calling the
  // startEditing() method and end by calling the 
  // finishEditing() method.
  // 
  // The startEditing() method has the side effect of returning
  // a pointer to the array of coordinates.
  //
  // Example:
  //
  //    SbVec3f *vertexPositions = coord->point.startEditing();
  //      for (ti .. )
  //        for (si .. )
  //          vertexPositions[index(si,ti)]. setValue( x,y,z );
  //    coord->point.finishEditing();
  //
  // Inventor automatically slices each rectangle of the
  // mesh into two triangles when the mesh is rendered.

  Grid.coord-> point.setNum(Grid.sDim*Grid.tDim);

  SbVec3f *vertexPositions = Grid.coord-> point.startEditing();

  Grid.mesh->verticesPerRow    = Grid.sDim;
  Grid.mesh->verticesPerColumn = Grid.tDim;

  sInit = StartingIndex % Grid.sDim;
  tInit = (StartingIndex / Grid.sDim) % Grid.tDim; 


  //
  // Main loop of the simulation.
  //
  // Traverse the grid and update the positions of the
  // vertices on the mesh.

  double k = 11.61984117, scale = 0.15;
  double dist;
  double dist2;
  angle = 2.0*M_PI * frequency * (SimulationTime - 0.0)  + phase;

  // x and y are [0,1]

  if (OptShowTimes){
    int clocktime = clock();
    printf("%d\n", clocktime-oldtime);
    oldtime=clocktime;
  }

  for (ti = tInit; (ti < Grid.tDim) && !breakFlag; ti++)
    {
    y = (Grid.tDim/2-ti)/(Grid.tDim-1.0);

    for (si = sInit; (si < Grid.sDim) && !breakFlag; si++){
      
      // This is the function z(x, y, t) being evaluated in the simulation.


      if (OptUseLookups){
	z = 0.0;
	for (d=0; d<MAX_DROPS && drops[d].time>=0.0; d++){
	  dist = L.lsdist(drops[d].s, drops[d].t, si, ti);
	  dist2 = dist-0.2*(SimulationTime-drops[d].time);      // speed of ring
          z += L.lnsin(dist*150.0-angle-1000.0)                 // tightness of rings
	    * exp(-dist2*dist2/(exp(-5.0*dist)*0.005))                 // width of ring
	    * exp(-dist*dist*2.0)
	    * scale;
	}
      }
      else{
	x = (Grid.sDim/2-si)/(Grid.sDim-1.0);
	z = scale*sin(sqrt(x*x+y*y)*50.0+angle);
      }
      
      // Put the newly-computed z-coordinate into the vertex (x,y,z).
      vertexPositions[ti*Grid.sDim + si][2] = z;

      gridpointsCompleted ++;
      breakFlag = (gridpointsCompleted >= gridpointBudget);
      }
    sInit = 0;
    }

  vertexPositions[0][2] = 5.0;
  Grid.coord->point.finishEditing();


  //
  // Update the grid asynchronously. 
  //
  // If the grid budget has been reached during the
  // updating of the mesh, return control to Inventor.
  // Then resume updating the grid on the next call
  // of updateMesh() . This leaves the grid in a 
  // partially updated state that the user can see.
  // Seeing a partially updated mesh may be considered 
  // objectionable by the user, but it allows
  // real-time interaction for moving the camera and
  // manipulating the draggers. 

  if (CompletedRows < Grid.tDim)
    {
    if (si == Grid.sDim)
      {
      CompletedRows = ti;
      }
    else
      {
      CompletedRows = ti-1;
      }
    Grid.mesh->verticesPerColumn = CompletedRows;
    }



    int completedGrid = ( (si == Grid.sDim) && (ti == Grid.tDim) );
    if (completedGrid)
      {
      SimulationTime += TimeIncrement;
      if ( (StartingIndex > 0) && (StartingIndex < Grid.sDim*Grid.tDim) &&
        (gridpointsCompleted < gridpointBudget) )
        {
	StartingIndex = 0;
	updateMesh(gridpointBudget - gridpointsCompleted);
	}
      else
        {
	StartingIndex = 0;
	}
      }
    else
      {
      StartingIndex = (StartingIndex + gridpointsCompleted);
      }

  }

